/****************************************************************************
Copyright 2010 Bilkent University

Licensed under the Apache License, Version 2.0 (the "License"); 
you may not use this file except in compliance with the License. 
You may obtain a copy of the License at 
http://www.apache.org/licenses/LICENSE-2.0 

Unless required by applicable law or agreed to in writing, software 
distributed under the License is distributed on an "AS IS" BASIS, 
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
See the License for the specific language governing permissions and 
limitations under the License. 
****************************************************************************/
#include "REng/GPU/GPUQuery.h"

#include <log4cplus/logger.h>
using namespace log4cplus;

#if RENG_GL_PLATFORM != RENG_GL_PLATFORM_ES

namespace REng{

	GLuint GPUQuery::ActiveQueryPrimsGenerated = 0;
	GLuint GPUQuery::ActiveQueryTransFeedback = 0;
	GLuint GPUQuery::ActiveQueryOcclusion = 0;
	GLuint GPUQuery::ActiveQueryTime = 0;

	GPUQuery::GPUQuery()
		:mIsActive(false){
		createResource();
	}
	GPUQuery::~GPUQuery() {
		destroyResource();
	}

	void GPUQuery::createResource(){
		glGenQueries(1,&mResourceID);
		assert(mResourceID!=0);
	}
	void GPUQuery::destroyResource(){
		glDeleteQueries(1,&mResourceID);
		mResourceID=0;
	}

	//////////////////////////////////////////////////////////////////////////
	// RESULT ACCESS

	bool GPUQuery::isActive() const{
		return mIsActive;
	}
	bool GPUQuery::isActive(QueryTarget target){
		switch(target){
			case QueryTarget_PrimsGenerated:
				return ActiveQueryPrimsGenerated!=0;
			case QueryTarget_TransFeedback: 
				return ActiveQueryTransFeedback!=0;
			case QueryTarget_SamplesPassed: 
			case QueryTarget_AnySamplesPassed:
				return ActiveQueryOcclusion!=0;
			case QueryTarget_TimeElapsed:
				return ActiveQueryTime!=0;
			default: assert(0); break;
		}
		return false;
	}
	bool GPUQuery::resultAvailable() const{
		if(mIsActive) {
			LOG4CPLUS_WARN(Logger::getInstance("RSys"),"Query result cannot be available on an active query");
			return false;
		}
		int p;
		glGetQueryObjectiv(mResourceID,GL_QUERY_RESULT_AVAILABLE,&p);
		assert(p==GL_TRUE || p==GL_FALSE);
		return (p==GL_TRUE);
	}
	void GPUQuery::getResultInt(int* p) const{
		if(mIsActive) return;
		assert(p);
		glGetQueryObjectiv(mResourceID,GL_QUERY_RESULT,p);
	}
	void GPUQuery::getResultUInt(uint* p) const{
		if(mIsActive) return;
		assert(p);
		glGetQueryObjectuiv(mResourceID,GL_QUERY_RESULT,p);
	}
	void GPUQuery::getResultBool(bool* b) const{
		if(mIsActive) return;
		int p;
		glGetQueryObjectiv(mResourceID,GL_QUERY_RESULT,&p);
		// TODO: OpenGL 3.3 and above supports direct boolean queries, we can use that directly
		*b = (p!=0);
	}

	void GPUQuery::setActiveQueryObject(QueryTarget target, GLuint objID){
		switch(target){
			case QueryTarget_PrimsGenerated:
				ActiveQueryPrimsGenerated = objID; break;
			case QueryTarget_TransFeedback: 
				ActiveQueryTransFeedback = objID; break;
			case QueryTarget_SamplesPassed: 
			case QueryTarget_AnySamplesPassed:
				ActiveQueryOcclusion = objID; break;
			case QueryTarget_TimeElapsed:
				ActiveQueryTime = objID; break;
			default: break;
		}
	}

	void GPUQuery::beginQuery(QueryTarget target){
		// check if ID is an active query name
		if(mIsActive) return;
#ifdef RENG_DEBUG_BUILD
		// ID is always generated by glGenQueries, deleted when object is destructed and cannot be zero.
		// ** Check if active query ID for target is non-zero
		switch(target){
			case QueryTarget_PrimsGenerated: 
				assert(ActiveQueryPrimsGenerated==0); break;
			case QueryTarget_TransFeedback: 
				assert(ActiveQueryTransFeedback==0); break;
			case QueryTarget_SamplesPassed:
			case QueryTarget_AnySamplesPassed:
				assert(ActiveQueryOcclusion==0); break;
			case QueryTarget_TimeElapsed:
				assert(ActiveQueryTime==0); break;
			default: break;
		}
		// ID always matches target (???)
		// TODO: check for conditional rendering state
#endif
		glBeginQuery(target,mResourceID);
		mIsActive = true;
		setActiveQueryObject(target,mResourceID);
	}

	void GPUQuery::endQuery(QueryTarget target){
#ifdef RENG_DEBUG_BUILD
		// We need to be de-activating the currently active query object
		// Note: It also makes sure that we have an active query for the given type
		switch(target){
			case QueryTarget_PrimsGenerated: 
				assert(ActiveQueryTransFeedback==mResourceID); 
				break;
			case QueryTarget_TransFeedback: 
				assert(ActiveQueryTransFeedback==mResourceID); 
				break;
			case QueryTarget_SamplesPassed: 
			case QueryTarget_AnySamplesPassed:
				assert(ActiveQueryOcclusion==mResourceID); 
				break;
			case QueryTarget_TimeElapsed:     
				assert(ActiveQueryTime==mResourceID); 
				break;
			default: break;
		}
#endif
		glEndQuery(target);
		mIsActive = false;
		setActiveQueryObject(target,0);
	}

	//////////////////////////////////////////////////////////////////////////
	// PRIMITIVE QUERIES
	//////////////////////////////////////////////////////////////////////////

	GPUQueryPrimGenerated::GPUQueryPrimGenerated(){
		begin();
		end();
	}
	void GPUQueryPrimGenerated::begin(){
		beginQuery(QueryTarget_PrimsGenerated);
	}
	void GPUQueryPrimGenerated::end(){
		endQuery(QueryTarget_PrimsGenerated);
	}
	void GPUQueryPrimTransFB::begin(){
		beginQuery(QueryTarget_TransFeedback);
	}
	void GPUQueryPrimTransFB::end(){
		endQuery(QueryTarget_TransFeedback);
	}

	//////////////////////////////////////////////////////////////////////////
	// OCCLUSION QUERIES
	//////////////////////////////////////////////////////////////////////////

	GPUQueryOcclusion::GPUQueryOcclusion(){
		begin();
		end();
	}
	void GPUQueryOcclusion::begin(){
		beginQuery(QueryTarget_SamplesPassed);
	}
	void GPUQueryOcclusion::beginAny(){
		beginQuery(QueryTarget_AnySamplesPassed);
	}
	void GPUQueryOcclusion::end(){
		endQuery(QueryTarget_SamplesPassed); // Note: QueryTarget_AnySamplesPassed follows same path
	}

	//////////////////////////////////////////////////////////////////////////
	// TIMER QUERIES
	//////////////////////////////////////////////////////////////////////////

	GPUQueryTimer::GPUQueryTimer(){
		begin();
		end();
	}
	void GPUQueryTimer::begin(){
		beginQuery(QueryTarget_TimeElapsed);
	}
	void GPUQueryTimer::end(){
		endQuery(QueryTarget_TimeElapsed);
	}
	void GPUQueryTimer::recordASynchGPUTime(){
		assert(!isActive());
		if(GLEW_VERSION_3_3 || GLEW_VERSION_4_0) {
			glQueryCounter(mResourceID,QueryTarget_Timestamp);
		}
	}
	int GPUQueryTimer::getSynchGPUTime() const{
		GLint t;
		glGetIntegerv(GL_TIMESTAMP,&t);
		return (int)t;
	}
	int64 GPUQueryTimer::getSynhGPUTime64() const{
		GLint64 t;
		glGetInteger64v(GL_TIMESTAMP,&t);
		return (int64)t;
	}

} // namespace REng

#endif // RENG_GL_PLATFORM != RENG_GL_PLATFORM_ES
